char *socketerror(int err) {
    switch (err) {
    case ECONNRESET:	return "Connection reset by peer";
			break;
    case ETIMEDOUT: 	return "Connection timed out.";
			break;
    case ECONNREFUSED: 	return "Connection refused.";
			break;
    case EHOSTDOWN: 	return "Host is down.";
			break;
    case EHOSTUNREACH: 	return "No route to host.";
			break;
    default:		return "Undefined error.";
    }
    return NULL;
}

struct in6_addr inet_addr6(char *addr) {
    struct in6_addr in6;
    unsigned int v6addr[8];
    int i;
    bzero(&in6, sizeof(struct in6_addr));
    if (!addr) return in6;
    if (!addr[0]) return in6;
    sscanf(addr, "%x:%x:%x:%x:%x:%x:%x:%x", &v6addr[0],
					    &v6addr[1],
					    &v6addr[2],
					    &v6addr[3],
					    &v6addr[4],
					    &v6addr[5],
					    &v6addr[6],
					    &v6addr[7]);
//    for (i=0; i < 8; i++) in6.s6_addr16[i]=htons(v6addr[i]);
    return in6;
}

void initnet() {
    int i;
    printf("*** Network ***\n");
    for (i=0; i < MAXSOCKETS; i++) socktab[i].type=SOCKTYPE_FREE;
    printf("Max sockets:     %d\n", MAXSOCKETS);
    printf("IPv6 support:    yes\n");
}

int socket_create(int type, int psid, Function onconnect, Function ondisconnect, Function onrecv, Function onclose, Function onerror) {
    int i, ret, fd, param;
    if (psid >= MAXSOCKETS) return (-1);
    for (i=0; i < MAXSOCKETS; i++) if (socktab[i].type == SOCKTYPE_FREE) {
	switch (type) {
	case SOCKTYPE_NORMAL:
	case SOCKTYPE_LISTEN:
	    ret=socket(AF_INET, SOCK_STREAM, 0);
	    break;
	case SOCKTYPE_NORMAL6:
	    ret=socket(AF_INET6, SOCK_STREAM, 0);
	    break;
	case SOCKTYPE_ACCEPT:
	    ret=0;
	    break;
	default:
 	    ret=(-1);
	    break;
	}
	if (ret < 0) return (-1);
	fd=ret;
	param=1; setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *) &param, sizeof(int));
	param=0; setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *) &param, sizeof(int));
	param=1; setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &param, sizeof(int));
	fcntl(fd, F_SETFL, O_NONBLOCK);
	bzero(&socktab[i], sizeof(struct ssock));
	socktab[i].type=type;
	socktab[i].fd=fd;
	socktab[i].psid=psid;
	socktab[i].onconnect=onconnect;
	socktab[i].ondisconnect=ondisconnect;
	socktab[i].onrecv=onrecv;
	socktab[i].onclose=onclose;
	socktab[i].onerror=onerror;
	socktab[i].inbuf=NULL;
	switch (type) {
	case SOCKTYPE_NORMAL:
	case SOCKTYPE_NORMAL6:
	case SOCKTYPE_ACCEPT:
	    socktab[i].inbuf=(char *) malloc(INBUFSIZE);
	    socktab[i].inbuf[0]=0;
	    break;
	default:
	    break;
	}
	return i;
    }
    return (-1);
}

int socket_bind(int sid, char *addr, int port) {
    struct sockaddr_in sin;
    struct sockaddr_in6 sin6;
    int ret, fd;
    if (sid >= MAXSOCKETS) return 0;
    if (socktab[sid].type == SOCKTYPE_FREE) return 0;
    if (!addr) return 0;
    if (!addr[0]) return 0;
    fd=socktab[sid].fd;
    switch (socktab[sid].type) {
    case SOCKTYPE_NORMAL:
        bzero((char *) &sin, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_addr.s_addr=inet_addr(addr);
	ret=bind(fd, (struct sockaddr *) &sin, sizeof(sin));
        break;
    case SOCKTYPE_LISTEN:
        bzero((char *) &sin, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(port);
	sin.sin_addr.s_addr=inet_addr(addr);
	ret=bind(fd, (struct sockaddr *) &sin, sizeof(sin));
        break;
    case SOCKTYPE_NORMAL6:
        bzero((char *) &sin6, sizeof(sin6));
	sin6.sin6_family=AF_INET6;
	sin6.sin6_addr=inet_addr6(addr);
	ret=bind(fd, (struct sockaddr *) &sin6, sizeof(sin6));
        break;
    default:
        ret=(-1);
        break;
    }
    if (ret < 0) return 0;
    return 1;
}

int socket_connect(int sid, char *addr, int port) {
    struct sockaddr_in sin;
    struct sockaddr_in6 sin6;
    int ret, fd;
    if (sid >= MAXSOCKETS) return 0;
    if (socktab[sid].type == SOCKTYPE_FREE) return 0;
    if (!addr) return 0;
    if (!addr[0]) return 0;
    fd=socktab[sid].fd;
    switch (socktab[sid].type) {
    case SOCKTYPE_NORMAL:
        bzero((char *) &sin, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(port);
	sin.sin_addr.s_addr=inet_addr(addr);
	memcpy(&socktab[sid].sin, &sin, sizeof(sin));
	ret=connect(fd, (struct sockaddr *) &sin, sizeof(sin));
        break;
    case SOCKTYPE_NORMAL6:
        bzero((char *) &sin6, sizeof(sin6));
	sin6.sin6_family=AF_INET6;
	sin6.sin6_port=htons(port);
	sin6.sin6_addr=inet_addr6(addr);
	memcpy(&socktab[sid].sin6, &sin6, sizeof(sin6));
	ret=connect(fd, (struct sockaddr *) &sin6, sizeof(sin6));
        break;
    default:
        return 0;
        break;
    }
    if ((ret < 0) && (errno != EINPROGRESS)) return 0;
    if (ret < 0) socktab[sid].status=SOCKNORMAL_CONNECT; 
    else {
	socktab[sid].status=SOCKNORMAL_CONNECTED;
	if (socktab[sid].onconnect) socktab[sid].onconnect((struct ssock *) &socktab[sid]);
    }
    return 1;
}

int socket_listen(int sid, int max) {
    int ret, fd;
    if (sid >= MAXSOCKETS) return 0;
    if (socktab[sid].type == SOCKTYPE_FREE) return 0;
    fd=socktab[sid].fd;
    switch (socktab[sid].type) {
    case SOCKTYPE_LISTEN:
	ret=listen(fd, max);
	socktab[sid].status=SOCKLISTEN_LISTEN;
        break;
    default:
        ret=(-1);
        break;
    }
    if (ret < 0) return 0;
    return 1;
}

int socket_accept(int psid) {
    struct sockaddr_in sin;
    int i, ret, fd, pfd, sid;
    if (psid >= MAXSOCKETS) return (-1);
    if (socktab[psid].type == SOCKTYPE_FREE) return (-1);
    pfd=socktab[psid].fd;
    switch (socktab[psid].type) {
    case SOCKTYPE_LISTEN:
	i=sizeof(sin);
	ret=accept(pfd, (struct sockaddr *) &sin, &i);
	if (ret < 0) return 0;
	fd=ret;
	ret=socket_create(SOCKTYPE_ACCEPT, psid, socktab[psid].onconnect,socktab[psid].ondisconnect, socktab[psid].onrecv, socktab[psid].onclose, socktab[psid].onerror);
	if (ret < 0) {
	    close(fd);
	    return 0;
	}
	sid=ret;
	memcpy(&socktab[sid].sin, &sin, sizeof(sin));
	socktab[sid].status=SOCKACCEPT_CONNECTED;
	socktab[sid].fd=fd;
	if (socktab[sid].onconnect) socktab[sid].onconnect((struct ssock *) &socktab[sid]);
        break;
    default:
        ret=(-1);
        break;
    }
    return ret;
}

int socket_closeid(int sid) {
    if (sid >= MAXSOCKETS) return 0;
    if (socktab[sid].type == SOCKTYPE_FREE) return 0;
    if (socktab[sid].onclose) socktab[sid].onclose((struct ssock *) &socktab[sid]);
    socktab[sid].type=SOCKTYPE_FREE;
    if (socktab[sid].inbuf) free(socktab[sid].inbuf);
    if (socktab[sid].fd >= 0) close(socktab[sid].fd);
    return 1;
}

int socket_closefd(int fd) {
    int i;
    for (i=0; i < MAXSOCKETS; i++) 
	if ((socktab[i].type != SOCKTYPE_FREE) && (socktab[i].fd == fd)) return socket_closeid(i);
    return 0;
}

int socket_idle() {
    fd_set fds;
    struct timeval t;
    int i, l, x, fdcnt, ret;
    char c;
    fdcnt=getdtablesize();
    
    FD_ZERO(&fds);
    for (i=0; i < MAXSOCKETS; i++) 
	if ((socktab[i].type != SOCKTYPE_FREE) && socktab[i].status) FD_SET(socktab[i].fd, &fds);
    t.tv_sec=0;
    t.tv_usec=1000;
    x=select(fdcnt, NULL, &fds, NULL, &t);
    if (x > 0) {
//	printf("1: x > 0\n");
        for (i=0; i < MAXSOCKETS; i++) 
	    if ((socktab[i].type != SOCKTYPE_FREE) && socktab[i].status) {
		if (FD_ISSET(socktab[i].fd, &fds)) {
		    switch (socktab[i].type) {
		    case SOCKTYPE_NORMAL:
		    case SOCKTYPE_NORMAL6:
			if (socktab[i].status == SOCKNORMAL_CONNECT) {
			    ret=recv(socktab[i].fd, &c, 1, MSG_PEEK);
			    if ((ret < 0) && (errno != EAGAIN)) {
				socktab[i].status=SOCKNORMAL_CONNECTERROR;
				socktab[i].errorno=errno;
				printf("(%d, %d) Unenable to connect, (%d) %s\n", i, socktab[i].fd, errno, socketerror(errno));
				if (socktab[i].onerror) socktab[i].onerror((struct ssock *) &socktab[i]);
			    } else {
				socktab[i].status=SOCKNORMAL_CONNECTED;
				printf("(%d, %d) Connected.\n", i, socktab[i].fd);
				if (socktab[i].onconnect) socktab[i].onconnect((struct ssock *) &socktab[i]);
			    }
			}
			break;
		    default:
    			break;
		    }
		}
	    }
    }

    t.tv_sec=0;
    t.tv_usec=100000;
    FD_ZERO(&fds);
    for (i=0; i < MAXSOCKETS; i++) 
	if ((socktab[i].type != SOCKTYPE_FREE) && socktab[i].status) FD_SET(socktab[i].fd, &fds);
    x=select(fdcnt, &fds, NULL, NULL, &t);
    if (x > 0) {
//	printf("2: x > 0\n");
        for (i=0; i < MAXSOCKETS; i++) 
	    if ((socktab[i].type != SOCKTYPE_FREE) && socktab[i].status) {
		if (FD_ISSET(socktab[i].fd, &fds)) {
		    switch (socktab[i].type) {
		    case SOCKTYPE_NORMAL:
		    case SOCKTYPE_NORMAL6:
			if (socktab[i].status == SOCKNORMAL_CONNECTED) {
			    l=strlen(socktab[i].inbuf);
			    if (INBUFSIZE-l) {
				ret=recv(socktab[i].fd, &socktab[i].inbuf[l], INBUFSIZE-l-1, 0);
				if (ret <= 0) {
				    socktab[i].status=SOCKNORMAL_DISCONNECTED;
				    socktab[i].errorno=errno;
				    printf("(%d, %d) EOF, %s\n", i, socktab[i].fd, socketerror(errno));
				    close(socktab[i].fd);
				    socktab[i].fd=(-1);
				    if (socktab[i].ondisconnect) socktab[i].ondisconnect((struct ssock *) &socktab[i]);
				    socket_closeid(i);
				} else if (ret > 0) {
				    socktab[i].inbuf[l+ret]=0;
				    if (socktab[i].onrecv) socktab[i].onrecv((struct ssock *) &socktab[i]);
				}
				if ((l+ret) > 14000) printf("*** Buffer overrun ***\n");
			    }
			}
			break;
		    case SOCKTYPE_ACCEPT:
			if (socktab[i].status == SOCKACCEPT_CONNECTED) {
			    l=strlen(socktab[i].inbuf);
			    if (INBUFSIZE-l) {
				ret=recv(socktab[i].fd, &socktab[i].inbuf[l], INBUFSIZE-l-1, 0);
				if (ret <= 0) {
				    socktab[i].status=SOCKACCEPT_DISCONNECTED;
				    socktab[i].errorno=errno;
				    printf("(%d, %d) EOF, %s\n", i, socktab[i].fd, socketerror(errno));
				    close(socktab[i].fd);
				    socktab[i].fd=(-1);
				    if (socktab[i].ondisconnect) socktab[i].ondisconnect((struct ssock *) &socktab[i]);
				    socket_closeid(i);
				} else if (ret > 0) {
				    socktab[i].inbuf[l+ret]=0;
				    if (socktab[i].onrecv) socktab[i].onrecv((struct ssock *) &socktab[i]);
				}
			    }
			}
			break;
		    case SOCKTYPE_LISTEN:
			socket_accept(i);
    			break;
		    default:
    			break;
		    }
		}
	    }
    }
}

void dprintf(int fd, char *f, ...) {
    char *buf;
    va_list va;
    va_start(va, f);
    vasprintf(&buf, f, va);
    va_end(va);
    send(fd, buf, strlen(buf), 0);
    free(buf);
}
